/* LZ77B-DEC.C - Decoding: LZ77 compression, non-adaptive arithmetic coding. */

#include "lz77.h"
#include "arithmetic_coding.h"
#include "freq.h"

/* Cumulative frequency tables for arithmetic coding. */

int match_cum_freq[Max_match+2];
int pos_cum_freq[Max_back+1];
int char_cum_freq[256+1];

#define NOT_EOF_symbol 1
#define EOF_symbol 2

int eof_cum_freq[2+1] = { 10000, 1, 0 };


/* Main program. */

int main(void)
{
  int i, m, p, r, c;
  int follow;

  /* Initialize for arithmetic decoding. */

  start_inputing_bits();
  start_decoding();

  /* Set up match length, position, following character tables for equal
     frequencies. */

  initialize_cum_freq(Max_match+1,match_cum_freq);
  initialize_cum_freq(Max_back,pos_cum_freq);
  initialize_cum_freq(256,char_cum_freq);

  for (;;)
  {
    /* Decode length and position of next match. */

    m = decode_symbol(match_cum_freq) - 1;
    if (m!=0) p = decode_symbol(pos_cum_freq);

    /* Output characters for this match, and also put them in "ahead". */

    for (i = m-1; i>=0; i--) 
    { c = back[Back_size-p-i];
      putchar(c);
      ahead[m-1-i] = c;
    }
  
    if (Debug)
    { fprintf(stderr,"Fetch m=%d p=%d '",m,p);
      for (i = m-1; i>=0; i--)
        if(back[Back_size-p-i]<32) fprintf(stderr,"\\%02x",back[Back_size-p-i]);
        else fprintf(stderr,"%c",back[Back_size-p-i]);
      fprintf(stderr,"'\n");
    }

    /* Decode the following character, unless we've disable it for this length 
       match (never disabled for standard scheme).  Also put in "ahead".  (But
       we first decode an EOF indicator flag, and terminate if EOF.) */

    follow = (send_follow>>m) & 1;

    if (follow)
    { 
      if (decode_symbol(eof_cum_freq) == EOF_symbol) break;

      c = decode_symbol(char_cum_freq) - 1;

      putchar(c);
      ahead[m] = c;
      if (Debug)
      { fprintf(stderr,"Extra: '");
        if (ahead[m]<32) fprintf(stderr,"\\%02x",ahead[m]); 
        else fprintf(stderr,"%c",ahead[m]);
        fprintf(stderr,"'\n");
      }
    }

    /* Shift characters from "ahead" into the "back" buffer. */

    shift(m+follow);
  }

  return 0;
}
