/* LZ77B-ENC.C - Encoding: LZ77 compression, non-adaptive arithmetic coding. */

#include "lz77.h"
#include "arithmetic_coding.h"
#include "freq.h"

/* Cumulative frequency tables for arithmetic coding. */

int match_cum_freq[Max_match+2];
int pos_cum_freq[Max_back+1];
int char_cum_freq[256+1];

#define NOT_EOF_symbol 1
#define EOF_symbol 2

int eof_cum_freq[2+1] = { 10000, 1, 0 };


/* Main program. */

int main(void)
{
  int i, m, p;
  int follow;

  /* Initialize for arithmetic encoding. */

  start_outputing_bits();
  start_encoding();

  /* Set up match length, position, following character tables for equal
     frequencies. */

  initialize_cum_freq(Max_match+1,match_cum_freq);
  initialize_cum_freq(Max_back,pos_cum_freq);
  initialize_cum_freq(256,char_cum_freq);

  /* Read characters (or EOFs) into "ahead" to get started. */

  for (i = 0; i<=Max_match; i++) ahead[i] = getchar();

  /* Send triples of (length, pos, char) until we reach EOF. */

  for (;;)
  {
    /* Find longest match.  Will find a length 0 match if we're at EOF. */

    match(&m,&p);

    /* Send length of match, and position if not length 0. */

    encode_symbol(m+1,match_cum_freq);
    if (m!=0) encode_symbol(p,pos_cum_freq);

    /* Send the following character, unless we're at end-of-file, or we've
       disable it for this length match (never disabled for standard scheme). 
       The following character is preceeded by an EOF indicator flag. */

    follow = (send_follow>>m) & 1;

    if (follow)
    { 
      encode_symbol (ahead[m]==EOF ? EOF_symbol : NOT_EOF_symbol, eof_cum_freq);

      if (ahead[m]==EOF) break;

      encode_symbol (ahead[m]+1, char_cum_freq);

      if (Debug)
      { fprintf(stderr,"Extra: '");
        if (ahead[m]<32) fprintf(stderr,"\\%02x",ahead[m]); 
        else fprintf(stderr,"%c",ahead[m]);
        fprintf(stderr,"'\n");
      }
    }

    /* Shift characters from the "ahead" buffer into the "back" buffer. */

    shift(m+follow);

    /* Shift characters down in the "ahead" buffer, and read new input 
       characters into the end. */

    for (i = 0; i<=Max_match-m-follow; i++) ahead[i] = ahead[i+m+follow];
    for ( ; i<=Max_match; i++) ahead[i] = getchar();
  }

  /* Finish with arithmetic encoding. */

  done_encoding();
  done_outputing_bits();

  return 0;
}
